/**
 * \file gtest_factor_gnss_single_diff_2D.cpp
 *
 *  Created on: Aug 28, 2018
 *      \author: jvallve
 */


#include "gnss/factor/factor_gnss_single_diff_2D.h"
#include <core/utils/utils_gtest.h>

#include "core/capture/capture_motion.h"
#include "core/problem/problem.h"
#include "gnss/sensor/sensor_gnss.h"
#include "core/sensor/sensor_odom_2D.h"
#include "core/processor/processor_odom_2D.h"
#include "gnss/processor/processor_gnss_single_diff.h"

#include "core/ceres_wrapper/ceres_manager.h"


using namespace Eigen;
using namespace wolf;
using std::cout;
using std::endl;


class FactorGnssSingleDiff2DTest : public testing::Test
{
    public:

        // groundtruth transformations
        Vector3s t_ecef_enu;
        Matrix3s R_ecef_enu, R_enu_map, R_map_base1, R_map_base2;
        Vector3s t_base_antena, t_ecef_antena1, t_ecef_antena2;
        Vector1s o_enu_map;
        Vector3s t_map_base1, t_map_base2;
        Vector1s o_map_base1, o_map_base2;

        // WOLF
        ProblemPtr problem_ptr;
        CeresManagerPtr ceres_mgr_ptr;
        SensorGnssPtr gnss_sensor_ptr;
        SensorOdom2DPtr odom_sensor_ptr;
        FrameBasePtr prior_frame_ptr;

        FactorGnssSingleDiff2DTest()
        {
            o_enu_map << 2.6;
            t_map_base1 << 32, -34, 0; // z=0
            o_map_base1 << -0.4;
            t_map_base2 << -76, 63, 0; // z=0
            o_map_base2 << 0.7;
            t_base_antena << 3,-2,8;// Antena extrinsics
            t_ecef_enu << 100,30,50;
            R_ecef_enu = (AngleAxis<Scalar>(1.3, Vector3s::UnitX())
                         *AngleAxis<Scalar>(-2.2, Vector3s::UnitY())
                         *AngleAxis<Scalar>(-1.8, Vector3s::UnitZ())).toRotationMatrix();

            // ----------------------------------------------------
            // Problem and solver
            problem_ptr = Problem::create("PO", 2);
            ceres_mgr_ptr = std::make_shared<CeresManager>(problem_ptr);
            ceres_mgr_ptr->getSolverOptions().max_num_iterations = 10;

            // gnss sensor
            gnss_sensor_ptr = std::static_pointer_cast<SensorGnss>(problem_ptr->installSensor("GNSS", "gnss", t_base_antena, std::make_shared<IntrinsicsBase>()));
            gnss_sensor_ptr->getEnuMapRoll()->fix();
            gnss_sensor_ptr->getEnuMapPitch()->fix();
            gnss_sensor_ptr->setEnuEcef(R_ecef_enu.transpose(), R_ecef_enu.transpose()*(-t_ecef_enu));

            // gnss processor
            ProcessorParamsBasePtr gnss_params_ptr = std::make_shared<ProcessorParamsBase>(true, 1.0);
            problem_ptr->installProcessor("GNSS SINGLE DIFFERENCES", "gnss single difference", gnss_sensor_ptr, gnss_params_ptr);

            // odom sensor & processor
            IntrinsicsOdom2DPtr odom_intrinsics_ptr = std::make_shared<IntrinsicsOdom2D>();
            odom_intrinsics_ptr->k_disp_to_disp = 0.1;
            odom_intrinsics_ptr->k_rot_to_rot = 0.1;
            odom_sensor_ptr = std::static_pointer_cast<SensorOdom2D>(problem_ptr->installSensor("ODOM 2D", "odometer", VectorXs::Zero(3), odom_intrinsics_ptr));
            ProcessorParamsOdom2DPtr odom_params_ptr = std::make_shared<ProcessorParamsOdom2D>();
            odom_params_ptr->voting_active = true;
            odom_params_ptr->dist_traveled = 1;
            odom_params_ptr->angle_turned = 2.0;
            odom_params_ptr->max_time_span = 1.0;
            odom_params_ptr->time_tolerance = 1.0;
            problem_ptr->installProcessor("ODOM 2D", "main odometry", odom_sensor_ptr, odom_params_ptr);
            //problem_ptr->setProcessorMotion("main odometry");

            // set prior (FIXED)
            Vector3s frame1state = t_map_base1;
            frame1state(2) = o_map_base1(0);
            prior_frame_ptr = problem_ptr->setPrior(frame1state, Matrix3s::Identity(), TimeStamp(0), Scalar(0.1));
            prior_frame_ptr->fix();
        };

        virtual void SetUp()
        {
            // reset grountruth parameters
            R_enu_map = AngleAxis<Scalar>(o_enu_map(0), Vector3s::UnitZ()).toRotationMatrix();
            R_map_base1 = Matrix3s::Identity();
            R_map_base1.topLeftCorner(2,2) = Rotation2D<Scalar>(o_map_base1(0)).matrix();
            R_map_base2 = Matrix3s::Identity();
            R_map_base2.topLeftCorner(2,2) = Rotation2D<Scalar>(o_map_base2(0)).matrix();

            t_ecef_antena1 = R_ecef_enu * R_enu_map * (R_map_base1 * t_base_antena + t_map_base1) + t_ecef_enu;
            t_ecef_antena2 = R_ecef_enu * R_enu_map * (R_map_base2 * t_base_antena + t_map_base2) + t_ecef_enu;

            // Reset antena extrinsics
            gnss_sensor_ptr->getP()->setState(t_base_antena);

            // Reset ENU_ECEF
            gnss_sensor_ptr->setEnuEcef(R_ecef_enu.transpose(), R_ecef_enu.transpose()*(-t_ecef_enu));
        }
};

////////////////////////////////////////////////////////

TEST_F(FactorGnssSingleDiff2DTest, check_tree)
{
    ASSERT_TRUE(problem_ptr->check(0));
}

/*
 * Test with one GNSS Single Difference capture.
 *
 * Estimating: MAP_BASE2 position.
 * Fixed: MAP_BASE1, ENU_MAP fixed, MAP_BASE2 orientation, BASE_ANTENA.
 */
TEST_F(FactorGnssSingleDiff2DTest, gnss_1_map_base_position)
{
    // Create GNSS Fix capture
    CaptureGnssSingleDiffPtr cap_gnss_ptr = std::make_shared<CaptureGnssSingleDiff>(TimeStamp(0), gnss_sensor_ptr, t_ecef_antena2-t_ecef_antena1, 1e-6*Matrix3s::Identity(), prior_frame_ptr);
    gnss_sensor_ptr->process(cap_gnss_ptr);

    // fixing things
    gnss_sensor_ptr->setEnuMapYawState(o_enu_map(0)); // enu-map yaw orientation
    gnss_sensor_ptr->getEnuMapYaw()->fix();
    cap_gnss_ptr->getFrame()->getO()->setState(o_map_base2); // frame orientation
    cap_gnss_ptr->getFrame()->getO()->fix();

    std::cout << "frame1: " << prior_frame_ptr->getState().transpose() << std::endl;

    // distort frm position
    Vector3s frm_dist = cap_gnss_ptr->getFrame()->getState();
    frm_dist(0) += 18;
    frm_dist(1) += -58;
    cap_gnss_ptr->getFrame()->setState(frm_dist);

    // solve for frm
    std::string report = ceres_mgr_ptr->solve(SolverManager::ReportVerbosity::BRIEF);

    ASSERT_TRUE(cap_gnss_ptr->getFrame()->isKey());
    ASSERT_MATRIX_APPROX(cap_gnss_ptr->getFrame()->getState().head(2), t_map_base2.head(2), 1e-6);
}

/*
 * Test with one GNSS Single Difference capture.
 *
 * Estimating: MAP_BASE2 orientation.
 * Fixed: MAP_BASE1, ENU_MAP fixed, MAP_BASE2 position, BASE_ANTENA.
 */
TEST_F(FactorGnssSingleDiff2DTest, gnss_1_map_base_orientation)
{
    // Create GNSS Fix capture
    CaptureGnssSingleDiffPtr cap_gnss_ptr = std::make_shared<CaptureGnssSingleDiff>(TimeStamp(0), gnss_sensor_ptr, t_ecef_antena2-t_ecef_antena1, 1e-6*Matrix3s::Identity(), prior_frame_ptr);
    gnss_sensor_ptr->process(cap_gnss_ptr);

    // fixing things
    gnss_sensor_ptr->setEnuMapYawState(o_enu_map(0)); // enu-map yaw orientation
    gnss_sensor_ptr->getEnuMapYaw()->fix();
    cap_gnss_ptr->getFrame()->getP()->setState(t_map_base2.head(2)); // frame position
    cap_gnss_ptr->getFrame()->getP()->fix();

    // distort frm position
    Vector1s frm_dist = cap_gnss_ptr->getFrame()->getO()->getState();
    frm_dist(0) += 1.8;
    cap_gnss_ptr->getFrame()->getO()->setState(frm_dist);

    // solve for frm
    std::string report = ceres_mgr_ptr->solve(SolverManager::ReportVerbosity::QUIET);

    ASSERT_TRUE(cap_gnss_ptr->getFrame()->isKey());
    ASSERT_MATRIX_APPROX(cap_gnss_ptr->getFrame()->getO()->getState(), o_map_base2, 1e-6);
}

/*
 * Test with one GNSS Single Difference capture.
 *
 * Estimating: ENU_MAP yaw.
 * Fixed: MAP_BASE1, MAP_BASE2 and BASE_ANTENA.
 */
TEST_F(FactorGnssSingleDiff2DTest, gnss_1_enu_map_yaw)
{
    // Create GNSS Fix capture
    CaptureGnssSingleDiffPtr cap_gnss_ptr = std::make_shared<CaptureGnssSingleDiff>(TimeStamp(0), gnss_sensor_ptr, t_ecef_antena2-t_ecef_antena1, 1e-6*Matrix3s::Identity(), prior_frame_ptr);
    gnss_sensor_ptr->process(cap_gnss_ptr);

    // unfixing things
    gnss_sensor_ptr->getEnuMapYaw()->unfix();
    // fixing things
    cap_gnss_ptr->getFrame()->getP()->setState(t_map_base2.head(2)); // frame position
    cap_gnss_ptr->getFrame()->getP()->fix();
    cap_gnss_ptr->getFrame()->getO()->setState(o_map_base2); // frame orientation
    cap_gnss_ptr->getFrame()->getO()->fix();

    // distort yaw enu_map
    Vector1s o_enu_map_dist = gnss_sensor_ptr->getEnuMapYaw()->getState();
    o_enu_map_dist(0) += 1.8;
    gnss_sensor_ptr->setEnuMapYawState(o_enu_map_dist(0));

    // solve for frm
    std::string report = ceres_mgr_ptr->solve(SolverManager::ReportVerbosity::QUIET);

    ASSERT_TRUE(cap_gnss_ptr->getFrame()->isKey());
    ASSERT_MATRIX_APPROX(gnss_sensor_ptr->getEnuMapYaw()->getState(), o_enu_map, 1e-6);
}

/*
 * Test with one GNSS Single Difference capture.
 *
 * Estimating: BASE_ANTENA (2 first components obsevable).
 * Fixed: MAP_BASE1, ENU_MAP, MAP_BASE2.
 */
TEST_F(FactorGnssSingleDiff2DTest, gnss_1_base_antena)
{
    // Create GNSS Fix capture
    CaptureGnssSingleDiffPtr cap_gnss_ptr = std::make_shared<CaptureGnssSingleDiff>(TimeStamp(0), gnss_sensor_ptr, t_ecef_antena2-t_ecef_antena1, 1e-6*Matrix3s::Identity(), prior_frame_ptr);
    gnss_sensor_ptr->process(cap_gnss_ptr);

    // unfixing things
    gnss_sensor_ptr->getP()->unfix();
    // fixing things
    gnss_sensor_ptr->setEnuMapYawState(o_enu_map(0)); // enu-map yaw orientation
    gnss_sensor_ptr->getEnuMapYaw()->fix();
    cap_gnss_ptr->getFrame()->getP()->setState(t_map_base2.head(2)); // frame position
    cap_gnss_ptr->getFrame()->getP()->fix();
    cap_gnss_ptr->getFrame()->getO()->setState(o_map_base2); // frame orientation
    cap_gnss_ptr->getFrame()->getO()->fix();

    // distort base_antena
    Vector3s base_antena_dist = gnss_sensor_ptr->getP()->getState();
    base_antena_dist(0) += 16.8;
    base_antena_dist(0) += -40.8;
    gnss_sensor_ptr->getP()->setState(base_antena_dist);

    // solve for frm
    std::string report = ceres_mgr_ptr->solve(SolverManager::ReportVerbosity::QUIET);

    ASSERT_TRUE(cap_gnss_ptr->getFrame()->isKey());
    ASSERT_MATRIX_APPROX(gnss_sensor_ptr->getP()->getState().head(2), t_base_antena.head(2), 1e-6);
}

int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}

